#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <cutil.h>
#include <cutil_gl_error.h>
#include <cuda_gl_interop.h>


#include <types.h>
#include <fhp_kernel.cu>

int num_of_particles;

TYPE c1[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE c2[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE c3[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE c4[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE c5[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE c6[LATTICE_WIDTH * LATTICE_HEIGHT];

TYPE k1[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE k2[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE k3[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE k4[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE k5[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE k6[LATTICE_WIDTH * LATTICE_HEIGHT];

TYPE boundary[LATTICE_WIDTH * LATTICE_HEIGHT];

TYPE *gpu_boundary;
TYPE *gpu_c1, *gpu_c2, *gpu_c3, *gpu_c4, *gpu_c5, *gpu_c6;
TYPE *gpu_k1, *gpu_k2, *gpu_k3, *gpu_k4, *gpu_k5, *gpu_k6;

dim3 block(BLOCK_DIM, BLOCK_DIM, 1);
dim3 grid(LATTICE_WIDTH / block.x, LATTICE_HEIGHT / block.y, 1);

float c1_avg, c2_avg, c3_avg, c4_avg, c5_avg, c6_avg;


int mouse_x, mouse_y;
unsigned int next_rand;

int random_int()
{
    next_rand = next_rand * 1103515245 + 12345;
    return (unsigned int) (next_rand / 65536);
}

int flip_coin()
{
    int r = random_int();
    return r % 2;
}

void init_cuda( )
{
	cudaMalloc((void **) &gpu_c1, ARRAY_LENGTH);
	cudaMalloc((void **) &gpu_c2, ARRAY_LENGTH);
	cudaMalloc((void **) &gpu_c3, ARRAY_LENGTH);
	cudaMalloc((void **) &gpu_c4, ARRAY_LENGTH);
	cudaMalloc((void **) &gpu_c5, ARRAY_LENGTH);
	cudaMalloc((void **) &gpu_c6, ARRAY_LENGTH);

	cudaMalloc((void **) &gpu_k1, ARRAY_LENGTH);
	cudaMalloc((void **) &gpu_k2, ARRAY_LENGTH);
	cudaMalloc((void **) &gpu_k3, ARRAY_LENGTH);
	cudaMalloc((void **) &gpu_k4, ARRAY_LENGTH);
	cudaMalloc((void **) &gpu_k5, ARRAY_LENGTH);
	cudaMalloc((void **) &gpu_k6, ARRAY_LENGTH);

	cudaMalloc((void **) &gpu_boundary, ARRAY_LENGTH);

	cudaMemcpy(gpu_c1, c1, ARRAY_LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(gpu_c2, c2, ARRAY_LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(gpu_c3, c3, ARRAY_LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(gpu_c4, c4, ARRAY_LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(gpu_c5, c5, ARRAY_LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(gpu_c6, c6, ARRAY_LENGTH, cudaMemcpyHostToDevice);
}

void next()
{
    /*
     *   3      2     *    c      b
     *    \    /      *     \    /
     *     \  /       *      \  /
     * 4 ---[]--- 1   *  d ---[]--- a
     *     /  \       *      /  \
     *    /    \      *     /    \
     *   5      6     *    e      f
     *
     */
	
	int i;	


	cudaMemcpy(gpu_c1, c1, ARRAY_LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(gpu_c2, c2, ARRAY_LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(gpu_c3, c3, ARRAY_LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(gpu_c4, c4, ARRAY_LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(gpu_c5, c5, ARRAY_LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(gpu_c6, c6, ARRAY_LENGTH, cudaMemcpyHostToDevice);

	cudaMemcpy(gpu_boundary, boundary, ARRAY_LENGTH, cudaMemcpyHostToDevice);

	/*
	cudaMemcpy(gpu_k1, k1, ARRAY_LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(gpu_k2, k2, ARRAY_LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(gpu_k3, k3, ARRAY_LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(gpu_k4, k4, ARRAY_LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(gpu_k5, k5, ARRAY_LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(gpu_k6, k6, ARRAY_LENGTH, cudaMemcpyHostToDevice);
	*/

	for(i = 0; i < FRAMES_PER_COMPUTATION; i++) {
		collision<<<grid, block>>>(gpu_c1, gpu_c2, gpu_c3, gpu_c4, gpu_c5, gpu_c6,
					   gpu_k1, gpu_k2, gpu_k3, gpu_k4, gpu_k5, gpu_k6,
					   gpu_boundary);	
	//	printf("before prop: c1[50]=%d, c2[50]=%d, c1[100]=%d\n", gpu_c1[50], gpu_c2[50], gpu_c1[100]);
		cudaThreadSynchronize( );
		propagate<<<grid, block>>>(gpu_c1, gpu_c2, gpu_c3, gpu_c4, gpu_c5, gpu_c6,
					   gpu_k1, gpu_k2, gpu_k3, gpu_k4, gpu_k5, gpu_k6,
					   gpu_boundary);
	//	printf("after prop: c1[50]=%d, c2[50]=%d, c1[100]=%d\n", gpu_c1[50], gpu_c2[50], gpu_c1[100]);
		cudaThreadSynchronize( );
	}

	cudaMemcpy(c1, gpu_c1, ARRAY_LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(c2, gpu_c2, ARRAY_LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(c3, gpu_c3, ARRAY_LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(c4, gpu_c4, ARRAY_LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(c5, gpu_c5, ARRAY_LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(c6, gpu_c6, ARRAY_LENGTH, cudaMemcpyDeviceToHost);
	//printf("after cpy: c1[50]=%d, c2[50]=%d, c1[100]=%d\n", c1[50], c2[50], c1[100]);

	/*
	cudaMemcpy(k1, gpu_k1, ARRAY_LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(k2, gpu_k2, ARRAY_LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(k3, gpu_k3, ARRAY_LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(k4, gpu_k4, ARRAY_LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(k5, gpu_k5, ARRAY_LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(k6, gpu_k6, ARRAY_LENGTH, cudaMemcpyDeviceToHost);
	*/

	//printf("c1[50]=%d, c2[50]=%d, c[2]=%d\n", c1[50], c2[50], c1[2]);
	//
	//cudaFree(gpu_c1);
	//cudaFree(gpu_c2);
	//cudaFree(gpu_c3);
	//cudaFree(gpu_c4);
	//cudaFree(gpu_c5);
	//cudaFree(gpu_c6);

	//cudaFree(gpu_k1);
	//cudaFree(gpu_k2);
	//cudaFree(gpu_k3);
	//cudaFree(gpu_k4);
	//cudaFree(gpu_k5);
	//cudaFree(gpu_k6);

	//cudaFree(gpu_boundary);
}

void draw_point(int x, int y, short value)
{
//	int t = value;
	if(value > 0) {
		glColor3f((float)value/(TYPE_LENGTH * 6.0), (float)value/(TYPE_LENGTH * 6.0), 0.9);
/*
		while(t > 0) {
			num_of_particles += (t & 0x1);
			t >>= 1;
		}
*/
		glVertex3f(x + WINDOW_WIDTH / 8, y + WINDOW_HEIGHT / 8, 0.0);
	}
}

void draw_line(float start_x, float start_y, float end_x, float end_y)
{
		glVertex3f(start_x + WINDOW_WIDTH/8, start_y + WINDOW_WIDTH/8, 0.0);	
		glVertex3f(end_x + WINDOW_WIDTH/8, end_y + WINDOW_WIDTH/8, 0.0);	
}

int num_of_ones(int x) {
	int t = 0;
	while(x) {
		t += x & 1;
		x >>= 1;
	}
	return t;
}

void display()
{
    int i, j;
    int x, y;
    float x_slope, y_slope; 
    num_of_particles = 0;
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glBegin(GL_POINTS);
    for (x = 0; x < LATTICE_WIDTH; x++) {
        for (y = 0; y < LATTICE_HEIGHT; y++) {
            i = x * SPACING;
            j = y * SPACING;

            if (boundary[y * LATTICE_WIDTH + x] == 0) {
                glColor3f(0.6, 0.2, 0.7);
                glVertex3f(i + WINDOW_WIDTH / 8, j + WINDOW_HEIGHT / 8, 0.0);
            }

/*
	    draw_point(i, j, c1[y * LATTICE_WIDTH + x] + c2[y * LATTICE_WIDTH + x] + c3[y * LATTICE_WIDTH + x] +
		     c4[y * LATTICE_WIDTH + x] + c5[y * LATTICE_WIDTH + x] + c6[y * LATTICE_WIDTH + x]);
*/


        }
    }
    glEnd();
    glBegin(GL_LINES);
    for (x = 0; x < LATTICE_WIDTH; x += VECTOR_FIELD_SPACING) {
        for (y = 0; y < LATTICE_HEIGHT; y += VECTOR_FIELD_SPACING) {
	    for(i = x; i < x + VECTOR_FIELD_SPACING; i++) {
		for(j = y; j < y + VECTOR_FIELD_SPACING; j++) {
			c1_avg += num_of_ones(c1[y * LATTICE_WIDTH + x]);
			c2_avg += num_of_ones(c2[y * LATTICE_WIDTH + x]);
			c3_avg += num_of_ones(c3[y * LATTICE_WIDTH + x]);
			c4_avg += num_of_ones(c4[y * LATTICE_WIDTH + x]);
			c5_avg += num_of_ones(c5[y * LATTICE_WIDTH + x]);
			c6_avg += num_of_ones(c6[y * LATTICE_WIDTH + x]);
		}
	    }
	    c1_avg /= (VECTOR_FIELD_SPACING * VECTOR_FIELD_SPACING * TYPE_LENGTH);
	    c2_avg /= (VECTOR_FIELD_SPACING * VECTOR_FIELD_SPACING * TYPE_LENGTH);
	    c3_avg /= (VECTOR_FIELD_SPACING * VECTOR_FIELD_SPACING * TYPE_LENGTH);
	    c4_avg /= (VECTOR_FIELD_SPACING * VECTOR_FIELD_SPACING * TYPE_LENGTH);
	    c5_avg /= (VECTOR_FIELD_SPACING * VECTOR_FIELD_SPACING * TYPE_LENGTH);
	    c6_avg /= (VECTOR_FIELD_SPACING * VECTOR_FIELD_SPACING * TYPE_LENGTH);

	    x_slope = c1_avg - c4_avg + SQRT_2*c2_avg - SQRT_2*c5_avg + SQRT_2*c6_avg - SQRT_2*c3_avg; 
	    y_slope = SQRT_2*c2_avg - SQRT_2*c5_avg - SQRT_2*c6_avg + SQRT_2*c3_avg; 
	    
//	    printf("x_slope is = %d, y_slope is = %d\n", x_slope, y_slope);
	    
	    glColor3f(1.0, 0.0, 0.0);
	    draw_line(x +  VECTOR_FIELD_SPACING * x_slope, y + VECTOR_FIELD_SPACING * y_slope, x, y);
        }
    }
    glEnd();

//    printf("Number of particles is %d\n", num_of_particles);

    glutSwapBuffers();
}

void idle()
{
	int i, j;
	if(random_int( ) % 113 == 0) {
    for (i = 0; i < LATTICE_HEIGHT / 20; i++) {
        for (j = 0; j < LATTICE_WIDTH / 20; j++) {
            if (1) {
                c1[i + LATTICE_WIDTH / 2 + LATTICE_WIDTH*(j + LATTICE_HEIGHT / 2)] = 0xF; //random_int( ) % 16;
                c2[i + LATTICE_WIDTH / 2 + LATTICE_WIDTH*(j + LATTICE_HEIGHT / 2)] = 0xF; //random_int( ) % 16;
                c3[i + LATTICE_WIDTH / 2 + LATTICE_WIDTH*(j + LATTICE_HEIGHT / 2)] = 0xF; //random_int( ) % 16;
                c4[i + LATTICE_WIDTH / 2 + LATTICE_WIDTH*(j + LATTICE_HEIGHT / 2)] = 0xF; //random_int( ) % 16;
                c5[i + LATTICE_WIDTH / 2 + LATTICE_WIDTH*(j + LATTICE_HEIGHT / 2)] = 0xF; //random_int( ) % 16;
                c6[i + LATTICE_WIDTH / 2 + LATTICE_WIDTH*(j + LATTICE_HEIGHT / 2)] = 0xF; //random_int( ) % 16;
            }
        }
    }
}
	/*
	for (i = 0; i < LATTICE_HEIGHT / 20; i++) {
        for (j = 0; j < LATTICE_WIDTH / 20; j++) {
            if (1) {
                c1[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = random_int( ) % (int) pow(2.0, TYPE_LENGTH);
                c2[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = random_int( ) % (int) pow(2.0, TYPE_LENGTH);
                c3[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = random_int( ) % (int) pow(2.0, TYPE_LENGTH);
                c4[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = random_int( ) % (int) pow(2.0, TYPE_LENGTH);
                c5[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = random_int( ) % (int) pow(2.0, TYPE_LENGTH);
                c6[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = random_int( ) % (int) pow(2.0, TYPE_LENGTH);
            }
        }
    }
	*/
/*    printf("boundary at i=%d  is %d\n", WINDOW_WIDTH/3, boundary[WINDOW_WIDTH/3][0]);*/
    next();

    glutPostRedisplay();
}

void init()
{
    int i, j;
    next_rand = 0;
    init_cuda( );

    for (i = 0; i < LATTICE_WIDTH; i++) {
        for (j = 0; j < LATTICE_HEIGHT; j++) {
            c1[j * LATTICE_WIDTH + i] = 0;
            c2[j * LATTICE_WIDTH + i] = 0;
            c3[j * LATTICE_WIDTH + i] = 0;
            c4[j * LATTICE_WIDTH + i] = 0;
            c5[j * LATTICE_WIDTH + i] = 0;
            c6[j * LATTICE_WIDTH + i] = 0;
/*
	    if(random_int( ) % 11 == 0) {
                c1[j * LATTICE_WIDTH + i] = 0xF; //(int) pow(2.0, TYPE_LENGTH);
                c2[j * LATTICE_WIDTH + i] = 0xF; //(int) pow(2.0, TYPE_LENGTH);
                c3[j * LATTICE_WIDTH + i] = 0xF; //(int) pow(2.0, TYPE_LENGTH);
                c4[j * LATTICE_WIDTH + i] = 0xF; //(int) pow(2.0, TYPE_LENGTH);
                c5[j * LATTICE_WIDTH + i] = 0xF; //(int) pow(2.0, TYPE_LENGTH);
                c6[j * LATTICE_WIDTH + i] = 0xF; //(int) pow(2.0, TYPE_LENGTH);
            }
*/
            boundary[j * LATTICE_WIDTH + i] = 0xF;
        }
    }
/*
    for (i = 0; i < LATTICE_WIDTH; i++) {
        boundary[(LATTICE_HEIGHT - 1) * LATTICE_WIDTH + i] = 0x0;
        boundary[i] = 0x0;
    }

    for (j = 0; j < LATTICE_HEIGHT; j++) {
        boundary[j * LATTICE_WIDTH] = 0x0;
        boundary[(j+1) * LATTICE_WIDTH - 1] = 0x0;
    }
*/
/*	
	for(k = 0; k < 10; k++) {
		for (i = 0; i < LATTICE_HEIGHT / 20; i++) {
			for (j = 0; j < LATTICE_WIDTH / 20; j++) {
				if (1) {
					c1[k * j * LATTICE_WIDTH + k * i] = random_int( ) % 16;
					c2[k * j * LATTICE_WIDTH + k * i] = random_int( ) % 16;
					c3[k * j * LATTICE_WIDTH + k * i] = random_int( ) % 16;
					c4[k * j * LATTICE_WIDTH + k * i] = random_int( ) % 16;
					c5[k * j * LATTICE_WIDTH + k * i] = random_int( ) % 16;
					c6[k * j * LATTICE_WIDTH + k * i] = random_int( ) % 16;
				}
			}
		}
	}
	
*/
/*
    for (i = 0; i < LATTICE_HEIGHT / 20; i++) {
        for (j = 0; j < LATTICE_WIDTH / 20; j++) {
            if (1) {
                c1[j * LATTICE_WIDTH/2 + i + LATTICE_WIDTH / 2] = random_int( ) % 16;
                c2[j * LATTICE_WIDTH/2 + i + LATTICE_WIDTH / 2] = random_int( ) % 16;
                c3[j * LATTICE_WIDTH/2 + i + LATTICE_WIDTH / 2] = random_int( ) % 16;
                c4[j * LATTICE_WIDTH/2 + i + LATTICE_WIDTH / 2] = random_int( ) % 16;
                c5[j * LATTICE_WIDTH/2 + i + LATTICE_WIDTH / 2] = random_int( ) % 16;
                c6[j * LATTICE_WIDTH/2 + i + LATTICE_WIDTH / 2] = random_int( ) % 16;
            }
        }
    }
*/
    for (i = 0; i < LATTICE_HEIGHT / 20; i++) {
        for (j = 0; j < LATTICE_WIDTH / 20; j++) {
            if (1) {
                c1[i + LATTICE_WIDTH / 2 + LATTICE_WIDTH*(j + LATTICE_HEIGHT / 2)] = 0xF; //random_int( ) % 16; //(int) pow(2.0, TYPE_LENGTH);
                c2[i + LATTICE_WIDTH / 2 + LATTICE_WIDTH*(j + LATTICE_HEIGHT / 2)] = 0xF; //random_int( ) % 16; //(int) pow(2.0, TYPE_LENGTH);
                c3[i + LATTICE_WIDTH / 2 + LATTICE_WIDTH*(j + LATTICE_HEIGHT / 2)] = 0xF; //random_int( ) % 16; //(int) pow(2.0, TYPE_LENGTH);
                c4[i + LATTICE_WIDTH / 2 + LATTICE_WIDTH*(j + LATTICE_HEIGHT / 2)] = 0xF; //random_int( ) % 16; //(int) pow(2.0, TYPE_LENGTH);
                c5[i + LATTICE_WIDTH / 2 + LATTICE_WIDTH*(j + LATTICE_HEIGHT / 2)] = 0xF; //random_int( ) % 16; //(int) pow(2.0, TYPE_LENGTH);
                c6[i + LATTICE_WIDTH / 2 + LATTICE_WIDTH*(j + LATTICE_HEIGHT / 2)] = 0xF; //random_int( ) % 16; //(int) pow(2.0, TYPE_LENGTH);
            }
        }
    }
    for (i = 0; i < LATTICE_WIDTH / 8; i++) {
        for (j = 0; j < LATTICE_HEIGHT / 8; j++) {
		boundary[i + 50 + LATTICE_WIDTH*(j + 150)] = 0x0;
		boundary[i + LATTICE_WIDTH/5 + 100 + LATTICE_WIDTH*(j+40)] = 0x0;
        }
    }
    for (i = 0; i < LATTICE_WIDTH / 6; i++) {
        for (j = 0; j < 10; j++) {
		boundary[i + 150 + LATTICE_WIDTH*(j + LATTICE_HEIGHT/20)] = 0x0;
		boundary[i + 150 + LATTICE_WIDTH*(j + 10)] = 0x0;
		boundary[i + 215 + LATTICE_WIDTH*(j + LATTICE_HEIGHT/4)] = 0x0;
		boundary[i + 215 + LATTICE_WIDTH*(j + LATTICE_HEIGHT/3)] = 0x0;
        }
    }
    glClearColor(0.0, 0.0, 0.0, 0.0);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0.0, WINDOW_WIDTH, 0.0, WINDOW_HEIGHT, -20.0, 20.0);

    glMatrixMode(GL_MODELVIEW);

    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glShadeModel(GL_SMOOTH);
    glPointSize(POINT_SIZE);
}

void mouse(int button, int state, int x, int y)
{
    mouse_x = x;
    mouse_y = y;
}

int main(int argc, char **argv)
{
    mouse_x = 0;
    mouse_y = 0;

    printf("Block size is %d\n", TYPE_LENGTH);
    printf("A total of %d vectors and %d lattices\n", ARRAY_LENGTH, ARRAY_LENGTH/6);

    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    glutInitWindowSize(WINDOW_WIDTH, WINDOW_HEIGHT);
    glutCreateWindow("fhp");
    glEnable(GL_DEPTH_TEST);
    init();
    glutDisplayFunc(display);
    glutIdleFunc(idle);
    glutMainLoop();
    return 0;
}
